#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
import logging
import time
from datetime import timedelta
from typing import Any

from mdo_engine.utilities.database import check_host_port

from .core import Connector
from .pipeline import Tree, set_output_scope
from .utils.database import get_database_config

# Set up logging
module_logger = logging.getLogger(__name__)


class ConnectorMenu:
    """Hub agnostic superclass for menu functions"""

    __metaclass__ = abc.ABCMeta

    @property
    @abc.abstractmethod
    def _hub_name(self):
        return "Should never get here"

    def get_available(self, core, project):
        """Return all the available modules found as plugins"""

        connector = self._get_connector(project)

        names = connector.get_available_interfaces(core, project)

        return names

    def activate(self, core, project, interface_name):
        """Move an interface into the pipelines list."""

        connector = self._get_connector(project)

        # Store an object of the interface
        connector.activate_interface(core, project, interface_name)

    def get_active(self, core, project):
        connector = self._get_connector(project)

        active_interface_names = connector.get_active_interface_names(
            core, project
        )

        return active_interface_names

    def get_current(self, core, project):
        connector = self._get_connector(project)

        interface_name = connector.get_current_interface_name(core, project)

        return interface_name

    def get_scheduled(self, core, project):
        connector = self._get_connector(project)

        interface_names = connector.get_scheduled_interface_names(core, project)

        return interface_names

    def get_completed(self, core, project):
        connector = self._get_connector(project)

        interface_names = connector.get_completed_interface_names(core, project)

        return interface_names

    def is_executable(
        self, core, project, interface_name, allow_unavailable=False
    ):
        connector = self._get_connector(project)

        core.unmask_states(project)
        result = connector.is_interface_executable(
            core, project, interface_name, allow_unavailable
        )

        return result

    def is_completed(self, core, project, interface_name):
        connector = self._get_connector(project)

        result = connector.is_interface_completed(core, project, interface_name)

        return result

    def _get_connector(self, project, hub_name=None):
        if hub_name is None:
            hub_name = self._hub_name

        simulation = project.get_simulation()
        hub_ids = simulation.get_hub_ids()

        if hub_ids is None or hub_name not in hub_ids:
            errStr = (
                'Hub name "{}" is not contained in the ' "project."
            ).format(hub_name)
            raise ValueError(errStr)

        result = Connector(hub_name)

        return result

    def _execute(
        self,
        core,
        project,
        interface_name,
        force_level=None,
        allow_unavailable=False,
        set_output_level=True,
    ):
        """Execute the given interface"""

        connector = self._get_connector(project)

        # Execute the interface
        connector.execute_interface(
            core,
            project,
            interface_name,
            level=force_level,
            allow_unavailable=allow_unavailable,
            set_output_level=set_output_level,
        )

    def _auto_execute(self, core, project, force_level=None):
        """Auto execute all interfaces in the Hub."""

        connector = self._get_connector(project)

        connector.auto_execute(core, project, force_level)


class ProjectMenu(ConnectorMenu):
    @property
    def _hub_name(self):
        return "project"

    def new_project(self, core, project_title):
        """Create a new project with a default simulation.

        Args:
          core (dtocean_core.core.Core): A Core object.
          project: (dtocean_core.core.Project): A Core object.
          project_title (str): The title of the project.

        Returns:
          dtocean_core.core.Project: A new Project object.

        """

        new_project = core.new_project(project_title)
        core.new_hub(new_project)

        self.activate(core, new_project, "System Type Selection")

        return new_project

    def initiate_pipeline(self, core, project):
        """Add the module and theme Hub objects to the core simulation.
        The input requirements of the "System Type Selection" interface
        must be met.

        Note:
          It is assumed that the core contains an active simulation at this
          stage.

        Args:
          core (dtocean_core.core.Core): A Core object.

        Returns:
          dtocean_core.core.Core: Updated copy of Core object.

        """

        self._execute(
            core, project, "System Type Selection", set_output_level=False
        )

        # Build the module and theme hubs
        core.new_hub(project)
        core.new_hub(project)

        # Load the top level data - check the status of the database.
        if project.get_database_credentials() is None:
            return

        self.activate(core, project, "Site and System Options")

        # Its necessary to load the site and system data from the database now
        # so a Tree is required.
        tree = Tree()

        options_branch = tree.get_branch(
            core, project, "Site and System Options"
        )
        options_branch.read_auto(core, project)

    def initiate_options(self, core, project):
        # Load the top level data - check the status of the database.
        if project.get_database_credentials() is None:
            return

        if "Site and System Options" not in self.get_active(core, project):
            errStr = (
                "The pipeline must be initialised before "
                "initalising the database options."
            )
            raise RuntimeError(errStr)

        self._execute(
            core,
            project,
            "Site and System Options",
            set_output_level=False,
        )
        self.activate(core, project, "Site Boundary Selection")
        self.activate(core, project, "Database Filtering Interface")

    def initiate_bathymetry(self, core, project):
        """Initiate selection of lease area for bathymetry filtering"""

        # Load the top level data - check the status of the database.
        if project.get_database_credentials() is None:
            return

        if "Site Boundary Selection" not in self.get_active(core, project):
            errStr = (
                "Database options must be initialised prior to "
                "initalising the bathymetry."
            )
            raise RuntimeError(errStr)

        self._execute(
            core, project, "Site Boundary Selection", set_output_level=False
        )

        self.activate(core, project, "Project Boundaries Interface")

    def initiate_filter(self, core, project):
        """Filter the database based on user selections."

        Args:
          core (dtocean_core.core.Core): A Core object.

        Returns:
          dtocean_core.core.Core: Updated copy of Core object.

        """

        # Load the top level data - check the status of the database.
        if project.get_database_credentials() is None:
            return

        if "Database Filtering Interface" not in self.get_active(core, project):
            errStr = (
                "Database options must be initialised prior to "
                "filtering the database."
            )
            raise RuntimeError(errStr)

        self._execute(
            core,
            project,
            "Database Filtering Interface",
            set_output_level=False,
        )

    def initiate_dataflow(self, core, project, _next_hub="modules"):
        """Add a marker level to indicate completion of the filtering steps and
        the beginning of preparing a simulation by collecting data. All modules
        and themes must be activated before calling this function"""

        if not core.has_data(project, "hidden.pipeline_active"):
            errStr = "Dataflow can not be initiated before the pipeline"
            raise RuntimeError(errStr)

        # Add an indicator variable for this stage
        core.add_datastate(
            project, identifiers=["hidden.dataflow_active"], values=[True]
        )

        # Create a level for resetting to pre-data collection point
        init_level = "{} {}".format(_next_hub, core._markers["initial"])
        core.register_level(project, init_level, _next_hub)


class ModuleMenu(ConnectorMenu):
    """A class for controlling actions related to the simulations such as
    choosing which modules and themes are to be run and executing modules and
    themes.

    """

    @property
    def _hub_name(self):
        return "modules"

    def activate(self, core, project, module_name):
        """Move a module into the pipelines list."""

        if core.has_data(project, "hidden.dataflow_active"):
            errStr = (
                "Modules can not be added after the dataflow has been "
                "initiated"
            )
            raise RuntimeError(errStr)

        super(ModuleMenu, self).activate(core, project, module_name)

    def is_executable(
        self,
        core,
        project,
        module_name,
        check_themes=True,
        allow_unavailable=False,
    ):
        result = super(ModuleMenu, self).is_executable(
            core, project, module_name, allow_unavailable
        )

        if not check_themes:
            return result

        theme_connector = self._get_connector(project, "themes")

        can_execute_themes = theme_connector.is_auto_executable(
            core, project, allow_unavailable=True
        )

        result = result & can_execute_themes

        return result

    def execute_current(
        self,
        core,
        project,
        execute_themes=True,
        allow_unavailable=False,
        log_execution_time=True,
    ):
        module_name = self.get_current(core, project)

        if module_name is None:
            errStr = "No interface is scheduled for execution."
            raise RuntimeError(errStr)

        # Check if execution is OK
        if not self.is_executable(
            core,
            project,
            module_name,
            check_themes=execute_themes,
            allow_unavailable=allow_unavailable,
        ):
            errStr = "Not all inputs for module {}".format(module_name)

            if execute_themes:
                theme_connector = self._get_connector(project, "themes")
                theme_names = theme_connector.get_scheduled_interface_names(
                    core, project
                )
                theme_str = ", ".join(theme_names)

                errStr = "{} or theme(s) {}".format(errStr, theme_str)

            errStr = "{} have been satisfied".format(errStr)

            raise RuntimeError(errStr)

        # Unmask any states
        core.unmask_states(project)

        logStr = "Executing module '{}'".format(module_name)
        module_logger.info(logStr)

        # Record the start time
        if log_execution_time:
            start_time = time.monotonic()

        self._execute(
            core, project, module_name, allow_unavailable=allow_unavailable
        )

        # Log the execution time
        if log_execution_time:
            end_time = time.monotonic()
            duration = timedelta(seconds=end_time - start_time)
            logStr = ("Module '{}' execution duration was: " "{}").format(
                module_name, duration
            )
            module_logger.info(logStr)

        if not execute_themes:
            return

        theme_connector = self._get_connector(project, "themes")

        if theme_connector.any_scheduled(core, project):
            # Run the themes just on local data. Mask all outputs then
            # unmask only those containing the module name.
            core.mask_states(
                project,
                search_str=core._markers["output"],
                no_merge=True,
                update_status=False,
            )
            core.unmask_states(
                project, search_str=module_name, update_status=False
            )

            local_level = "{} {}".format(module_name, core._markers["local"])
            theme_connector.auto_execute(
                core,
                project,
                local_level,
                register_level=False,
                allow_non_execution=True,
            )

            # Run the themes for global data. Unmask all the states and create
            # a modified level
            core.unmask_states(project)

            global_level = "{} {}".format(module_name, core._markers["global"])
            theme_connector.auto_execute(
                core,
                project,
                global_level,
                register_level=False,
                allow_non_execution=True,
            )

            # Set global scope for theme outputs
            set_output_scope(core, project)


class ThemeMenu(ConnectorMenu):
    """A class for controlling actions related to the simulations such as
    choosing which modules and themes are to be run and executing modules and
    themes.

    """

    @property
    def _hub_name(self):
        return "themes"

    def activate(self, core, project, interface_name):
        """Move a theme into the pipelines list."""

        if core.has_data(project, "hidden.dataflow_active"):
            errStr = (
                "Themes can not be added after the dataflow has been "
                "initiated"
            )
            raise RuntimeError(errStr)

        super(ThemeMenu, self).activate(core, project, interface_name)

    def execute_all(self, core, project):
        """Execute a theme

        Process
        -------

        1. Test if data requirements are met
        2. Add the inputs into the interface and execute
        3. Add the results to the data state

        """

        level = "themes"
        theme_connector = self._get_connector(project, "themes")

        # Unmask any states
        core.unmask_states(project)

        theme_connector.auto_execute(core, project, level)


class DataMenu:
    """The DataMenu class is for actions relating to populating the core with
    data from various sources.
    """

    def __init__(self):
        self._tree = Tree()
        self._useryaml, self._dbconfig = get_database_config()

    def get_available_databases(self):
        return list(self._dbconfig.keys())

    def get_database_dict(self, identifier):
        if identifier not in self._dbconfig:
            errStr = (
                "Identifier {} does not correspond to any found "
                "database credentials."
            ).format(identifier)
            raise ValueError(errStr)

        db_dict = self._dbconfig[identifier]

        return db_dict

    def update_database(self, identifier, db_dict: dict, allow_new=False):
        if not identifier:
            errStr = "Database identifier must be a valid string."
            raise ValueError(errStr)

        if not allow_new and identifier not in self._dbconfig:
            errStr = (
                "Identifier {} does not correspond to any found "
                "database credentials."
            ).format(identifier)
            raise ValueError(errStr)

        safe_db_dict = {}

        for k, v in db_dict.items():
            if not v:
                continue
            safe_db_dict[k] = v

        self._dbconfig[identifier] = safe_db_dict
        self._useryaml.write(self._dbconfig)

    def delete_database(self, identifier):
        if identifier not in self._dbconfig:
            errStr = (
                "Identifier {} does not correspond to any found "
                "database credentials."
            ).format(identifier)
            raise ValueError(errStr)

        self._dbconfig.pop(identifier, None)
        self._useryaml.write(self._dbconfig)

    def check_database(self, identifier=None, credentials=None):
        credentials = self._get_credentials(identifier, credentials)

        if "port" in credentials:
            port = credentials["port"]
        else:
            port = 5432

        port_open, msg = check_host_port(credentials["host"], port)
        module_logger.debug(msg)

        return port_open

    def select_database(self, project, identifier=None, credentials=None):
        credentials = self._get_credentials(identifier, credentials)

        if not self.check_database(credentials=credentials):
            if "dbname" in credentials:
                add_str = " for database '{}'".format(credentials["dbname"])
            else:
                add_str = ""

            err_str = "Port could not be opened" + add_str
            raise RuntimeError(err_str)

        project.set_database_credentials(credentials)

    def deselect_database(self, project):
        project.set_database_credentials(None)

    def load_data(self, core, project):
        if not core.has_data(project, "hidden.dataflow_active"):
            errStr = (
                "Bulk data can not be loaded until the dataflow has "
                "been initiated."
            )
            raise RuntimeError(errStr)

        # Load the main data - check the status of the database.
        if project.get_database_credentials() is None:
            return

        self._tree.read_auto(core, project, ("modules", "themes"))

    def export_data(self, core, project, file_path, mask_outputs=False):
        args = [project, file_path]

        if mask_outputs:
            args.append(core._markers["output"])

        core.dump_datastate(*args)

    def import_data(self, core, project, file_path, skip_satisfied=False):
        args = [project, file_path]
        kwargs: dict[str, Any] = {"exclude": "hidden"}

        if skip_satisfied:
            kwargs["overwrite"] = False

        core.load_datastate(*args, **kwargs)

    def _get_credentials(self, identifier=None, credentials=None):
        if credentials is None and identifier is None:
            error_str = (
                "Either 'identifier' or 'credentials' keywords must "
                "be passed"
            )
            raise ValueError(error_str)

        if credentials is None and identifier not in self._dbconfig:
            errStr = (
                "Identifier {} does not correspond to any found "
                "database credentials."
            ).format(identifier)
            raise ValueError(errStr)

        if credentials is None:
            credentials = self._dbconfig[identifier]

        return credentials



import warnings

from cartopy.feature import NaturalEarthFeature

from dtocean_core.utils.config import init_config
from dtocean_core.utils.execute import main

from .shared import SmartFormatter


def init(args):
    print("+ Downloading cartopy plot data...")

    # Automatically scaled country boundaries
    borders = NaturalEarthFeature(
        "cultural",
        "admin_0_boundary_lines_land",
        "50m",
    )

    # Automatically scaled coastline, including major islands
    coastline = NaturalEarthFeature(
        "physical",
        "coastline",
        "50m",
    )

    # Automatically scaled land polygons, including major islands.
    land = NaturalEarthFeature(
        "physical",
        "land",
        "50m",
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        borders.geometries()
        coastline.geometries()
        land.geometries()


def subcommand(subparser):
    parser = subparser.add_parser("core", help="dtocean-core utilities")
    sp = parser.add_subparsers(required=True)
    _setup_run(sp)
    _setup_config(sp)


def _setup_run(subparser):
    desStr = (
        "Execute DTOcean .dtop project files. By default, the next "
        "module scheduled is executed. All scheduled modules can also "
        "be run using the appropriate option. Completed simulations are "
        "saved to a new project file with '_complete' appended to the "
        "file path."
    )

    parser = subparser.add_parser(
        "run",
        description=desStr,
        help="run simulation",
    )

    parser.add_argument(
        "fpath", help=("path to DTOcean project file"), type=str
    )

    parser.add_argument(
        "-o",
        "--out",
        help=("specify output path for results file"),
        type=str,
        default=None,
    )

    parser.add_argument(
        "-f",
        "--full",
        help=("execute all scheduled modules"),
        action="store_true",
    )

    parser.add_argument(
        "-n", "--no-save", help=("do not save the results"), action="store_true"
    )

    parser.add_argument(
        "-w",
        "--warnings",
        help=("show tracebacks for all warnings"),
        action="store_true",
    )

    parser.add_argument(
        "-l",
        "--logging",
        help=("activate the DTOcean logging system"),
        action="store_true",
    )

    parser.set_defaults(func=lambda args: _main(args))


def _main(args):
    fpath = args.fpath
    rpath = args.out
    full = args.full
    warn = args.warnings
    no_save = args.no_save
    log = args.logging

    if no_save is True:
        save = False
    elif rpath is not None:
        save = rpath.strip()
    else:
        save = True

    main(fpath, save, full, warn, log)


def _setup_config(subparser):
    """Command line parser for init_config.

    Example:

        To get help::

            dtocean core config -h

    """

    desStr = (
        "Copy user modifiable configuration files to "
        r"<UserName>\AppData\Roaming\DTOcean\dtocean-core\config"
    )

    parser = subparser.add_parser(
        "config",
        description=desStr,
        help="copy config files",
        formatter_class=SmartFormatter,
    )

    parser.add_argument(
        "action",
        choices=["logging", "database"],
        help="R|Select an action, where\n"
        " logging = copy logging configuration\n"
        " database = copy database configuration",
    )

    parser.add_argument(
        "--overwrite",
        help=("overwrite any existing configuration files"),
        action="store_true",
    )

    parser.set_defaults(func=lambda args: _config(args))


def _config(args):
    action = args.action
    overwrite = args.overwrite

    kwargs = {
        "logging": False,
        "database": False,
        "overwrite": overwrite,
    }

    kwargs[action] = True
    dir_path = init_config(**kwargs)

    if dir_path is not None:
        print("Copying configuration files to {}".format(dir_path))



# -*- coding: utf-8 -*-

#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from copy import deepcopy
from typing import Any, Optional

from dtocean_core.menu import ModuleMenu, ThemeMenu
from dtocean_core.pipeline import Tree


class Strategy(object):
    """The base abstract class for all strategy classes"""

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """The init method should never have arguments. Provided for all
        subclasses are:

            self._module_menu: a ModuleMenu object
            self._theme_menu: a ThemeMenu object
            self._tree: a Tree object

        """

        self._module_menu = ModuleMenu()
        self._theme_menu = ThemeMenu()
        self._tree = Tree()

        # Record the simulation titles used in the strategy
        self._sim_record: list[str] = []

        # Record the configuration dictionary of the strategy
        self._config: Optional[dict[str, Any]] = None

        # Record any detailed information about the simulation (assume this
        # is json serializable)
        self.sim_details: Optional[Any] = None

    @property
    @abc.abstractmethod
    def version(self) -> int:
        """Version identifier for backwards compatibility when deserializing"""
        pass

    @classmethod
    @abc.abstractmethod
    def get_name(cls):
        """A class method for the common name of the strategy.

        Returns:
          str: A unique string
        """

        return cls()

    @abc.abstractmethod
    def configure(self):
        """The configure method is collect information required for executing
        the strategy.
        """

    @abc.abstractmethod
    def get_variables(self):
        """The get_variables method returns the list of any variables that
        will be set by the strategy
        """

    @abc.abstractmethod
    def execute(self, core, project):
        """The execute method is used to execute the strategy. It should always
        take a Core and a Project class as the only inputs.
        """

    @staticmethod
    @abc.abstractmethod
    def dump_config(config):
        """Convert the config in a json serializable state"""

    @staticmethod
    @abc.abstractmethod
    def load_config(serial_config: Optional[dict], version: int):
        """Convert the json representation of the config"""

    @staticmethod
    @abc.abstractmethod
    def dump_sim_details(sim_details):
        """Convert the sim details in a json serializable state"""

    @staticmethod
    @abc.abstractmethod
    def load_sim_details(serial_sim_details):
        """Convert the json representation of the sim details"""

    def get_config(self):
        return deepcopy(self._config)

    def set_config(self, config_dict):
        self._config = config_dict

    def add_simulation_title(self, sim_title):
        self._sim_record.append(sim_title)

    def remove_simulation_title(self, sim_title):
        self._sim_record.pop(self._sim_record.index(sim_title))

    def get_simulation_record(self):
        return self._sim_record[:]

    def restart(self):
        self._sim_record = []
        self.sim_details = None



#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import sys
from pathlib import Path
from pickle import load as load_pkl
from typing import Optional, Sequence, Union

import matplotlib.pyplot as plt
from mdo_engine.boundary.interface import FileInterface

from .core import Connector, Core

StrOrPath = Union[str, Path]

# Set up logging
module_logger = logging.getLogger(__name__)


class Tree:
    """Interface class for the pipeline at the tree level"""

    def get_available_branches(self, core, project, hub_names=None):
        hub_branch_names = self._get_branch_names(core, project, hub_names)
        all_branch_names = [y for x in hub_branch_names.values() for y in x]

        return all_branch_names

    def get_branch(self, core, project, branch_name):
        available_branches = self.get_available_branches(core, project)

        if branch_name not in available_branches:
            errStr = (
                "Branch name {} not in list of available branches."
            ).format(branch_name)
            raise ValueError(errStr)

        if available_branches.count(branch_name) > 1:
            errStr = ("Branch name {} is not unique.").format(branch_name)
            raise ValueError(errStr)

        # Find the branch and hub
        hub_branch_names = self._get_branch_names(core, project)
        result = None

        for hub_name, branch_names in hub_branch_names.items():
            if branch_name in branch_names:
                new_branch = Branch(hub_name, branch_name)
                result = new_branch
                break

        if result is None:
            errStr = "Branch not found. Evil forces are at play!"
            raise RuntimeError(errStr)

        return result

    def read(
        self,
        core,
        project,
        hub_names=None,
        overwrite=False,
        skip_unavailable=True,
        set_auto=False,
        log_exceptions=False,
    ):
        branch_names = self.get_available_branches(core, project, hub_names)
        branch_list = [
            self.get_branch(core, project, name) for name in branch_names
        ]

        # Work out the unique inputs across all branches.
        unique_branch_inputs = set()
        variables = []

        for branch in branch_list:
            branch_input_status = branch.get_input_status(core, project)
            if branch_input_status is None:
                continue
            assert isinstance(branch_input_status, dict)

            if skip_unavailable:
                branch_inputs = [
                    k
                    for k, v in branch_input_status.items()
                    if "unavailable" not in v
                ]
            else:
                branch_inputs = branch_input_status.keys()

            unique_branch_inputs.update(branch_inputs)

        for variable_id in unique_branch_inputs:
            new_variable = InputVariable(variable_id)
            variables.append(new_variable)

        _read_variables(
            core, project, variables, overwrite, set_auto, log_exceptions
        )

    def read_auto(
        self,
        core,
        project,
        hub_names=None,
        overwrite=False,
        skip_unavailable=True,
        log_exceptions=True,
    ):
        """Calls read with set_auto set to True."""

        self.read(
            core,
            project,
            hub_names,
            overwrite,
            skip_unavailable,
            True,
            log_exceptions,
        )

    def _get_branch_names(self, core, project, hub_names=None):
        all_branch_names = {}

        if hub_names is None:
            simulation = project.get_simulation()
            hub_names = simulation.get_hub_ids()

        for hub_name in hub_names:
            connector = _get_connector(project, hub_name)
            branch_names = connector.get_active_interface_names(core, project)
            all_branch_names[hub_name] = branch_names

        return all_branch_names


class Branch:
    """Interface class for the pipeline at the branch level"""

    def __init__(self, hub_name, name):
        self._hub_name = hub_name
        self._name = name

    def get_inputs(self, core, project):
        connector = _get_connector(project, self._hub_name)
        input_declaration = connector.get_interface_inputs(
            core, project, self._name
        )

        return input_declaration

    def get_outputs(self, core, project):
        connector = _get_connector(project, self._hub_name)
        output_declaration = connector.get_interface_outputs(
            core, project, self._name
        )

        return output_declaration

    def get_input_status(
        self,
        core,
        project,
        variable=None,
    ) -> Optional[dict | str]:
        connector = _get_connector(project, self._hub_name)

        branch_status = connector.get_interface_inputs_status(
            project, self._name
        )

        if branch_status is None:
            return None

        if variable is not None:
            if not isinstance(variable, InputVariable):
                errStr = (
                    "Argument 'variable' only accepts InputVariable " "objects."
                )
                raise ValueError(errStr)

            if variable._id not in branch_status:
                errStr = (
                    "Variable {} is not an input to branch " "{}."
                ).format(variable._id, self._name)
                raise ValueError(errStr)

            variable_status = branch_status[variable._id]

            return variable_status

        else:
            return branch_status

    def get_output_status(self, core, project, variable=None):
        connector = _get_connector(project, self._hub_name)

        branch_status = connector.get_interface_outputs_status(
            project, self._name
        )

        if variable is not None:
            if not isinstance(variable, OutputVariable):
                errStr = (
                    "Argument 'variable' only accepts OutputVariable "
                    "objects."
                )
                raise ValueError(errStr)

            if variable._id not in branch_status:
                errStr = (
                    "Variable {} is not an output of branch " "{}."
                ).format(variable._id, self._name)
                raise ValueError(errStr)

            variable_status = branch_status[variable._id]

            return variable_status

        else:
            return branch_status

    def get_input_variable(
        self,
        core,
        project,
        variable_id,
        skip_missing=False,
    ):
        core.check_valid_variable(variable_id)

        # Check that the inputs exist in the tree
        module_inputs = self.get_input_status(core, project)
        if module_inputs is None:
            return None

        assert isinstance(module_inputs, dict)

        if variable_id not in module_inputs.keys():
            msgStr = ("Variable {} is not an input to branch " "{}.").format(
                variable_id, self._name
            )

            if skip_missing:
                module_logger.debug(msgStr)
                return None

            raise KeyError(msgStr)

        new_variable = InputVariable(variable_id)

        return new_variable

    def get_output_variable(self, core, project, variable_id):
        core.check_valid_variable(variable_id)

        # Check that the inputs exist in the tree
        module_inputs = self.get_output_status(core, project)

        if variable_id not in module_inputs.keys():
            errStr = ("Variable {} is not an output of branch " "{}.").format(
                variable_id, self._name
            )

            raise KeyError(errStr)

        new_variable = OutputVariable(variable_id)

        return new_variable

    def inspect(self, core, project, scope=None):
        connector = _get_connector(project, self._hub_name)

        # Can't inspect unless the interface has executed.
        if not connector.is_interface_completed(core, project, self._name):
            return

        output_level = "{} {}".format(
            self._name.lower(), core._markers["output"]
        )
        register_level = "{} {}".format(
            self._name.lower(), core._markers["register"]
        )

        # Call inspect level on the output level, but force the inspection
        # level to the theme scope.
        core.inspect_level(
            project,
            output_level,
            inspection_level=register_level,
            update_status=False,
        )

        # Unmask the levels relevant to the chosen theme output scope.
        set_output_scope(core, project, scope)

    def reset(self, core, project):
        connector = _get_connector(project, self._hub_name)

        # Can't reset unless the interface has executed.
        if not connector.is_interface_completed(core, project, self._name):
            return

        registered_level = "{} {}".format(
            self._name.lower(), core._markers["register"]
        )
        core.reset_level(
            project,
            registered_level,
            force_scheduled="themes",
            skip_missing=True,
        )

    def read(
        self,
        core,
        project,
        overwrite=False,
        skip_unavailable=True,
        set_auto=False,
        log_exceptions=False,
    ):
        # Control bulk reads for post project hubs
        if self._hub_name in ["modules", "themes"] and not core.has_data(
            project, "hidden.dataflow_active"
        ):
            errStr = (
                "Bulk data can not be loaded until the dataflow has "
                "been initiated."
            )
            raise RuntimeError(errStr)

        branch_input_status = self.get_input_status(core, project)
        if branch_input_status is None:
            return None

        assert isinstance(branch_input_status, dict)
        variables = []

        for identifier, status in branch_input_status.items():
            if skip_unavailable and "unavailable" in status:
                continue

            var = self.get_input_variable(core, project, identifier)

            variables.append(var)

        _read_variables(
            core, project, variables, overwrite, set_auto, log_exceptions
        )

    def read_auto(
        self,
        core,
        project,
        overwrite=False,
        skip_unavailable=True,
        log_exceptions=True,
    ):
        """Calls read with set_auto set to True."""

        self.read(
            core, project, overwrite, skip_unavailable, True, log_exceptions
        )

    def read_test_data(
        self,
        core,
        project,
        data_path,
        overwrite=False,
        skip_unavailable=True,
        skip_missing=True,
    ):
        with open(data_path, "rb") as dataf:
            test_data = load_pkl(dataf)

        variables = []

        for variable_id, value in test_data.items():
            variable = self.get_input_variable(
                core, project, variable_id, skip_missing=skip_missing
            )

            if skip_unavailable:
                status = self.get_input_status(core, project, variable)
                if status is None:
                    continue

                if "unavailable" in status:
                    continue

            # Loop if no variable is found
            if variable is None:
                continue

            variable.set_raw_interface(core, value)
            variables.append(variable)

        _read_variables(core, project, variables, overwrite=overwrite)


class Variable:
    """Interface class for the pipeline at the variable level"""

    def __init__(self, identifier):
        self._id = identifier
        self._interface = None

    def get_metadata(self, core):
        # Get the meta data from the core data catalog
        metadata = core.get_metadata(self._id)

        return metadata

    def has_value(self, core, project):
        result = core.has_data(project, self._id)

        return result

    def get_value(self, core, project):
        data_value = core.get_data_value(project, self._id)

        return data_value

    def get_file_output_interfaces(self, core, project, include_auto=False):
        kargs = {
            "core": core,
            "project": project,
            "interface_name": "FileOutputInterface",
            "dict_value_attr": "get_valid_extensions",
        }
        if include_auto:
            kargs["auto_interface_name"] = "AutoFileOutput"

        provider_names = self._get_receivers(**kargs)

        return provider_names

    def get_available_plots(self, core, project, include_auto=False):
        args = [core, project, "PlotInterface"]
        if include_auto:
            args.append("AutoPlot")

        active_receivers = self._get_receivers(*args)
        assert isinstance(active_receivers, list)
        active_receivers.sort()

        return active_receivers

    def write_file(self, core, project, file_path, interface_name=None):
        interface = self._get_receiving_interface(
            core,
            project,
            "FileOutputInterface",
            "AutoFileOutput",
            interface_name,
        )

        if interface is None or not core.has_data(project, self._id):
            return

        interface.set_file_path(file_path)
        self._write_interface(core, project, interface)

    def plot(self, core, project, plot_name=None):
        interface = self._get_receiving_interface(
            core, project, "PlotInterface", "AutoPlot", plot_name
        )

        if interface is None or not core.has_data(project, self._id):
            return

        self._write_interface(core, project, interface)
        plt.show(block=False)

    def _get_named_interface(
        self, core, socket_cls_name, interface_name, allow_missing=False
    ):
        socket = core.socket_map[socket_cls_name]

        interface_name_map = socket.get_interface_names()

        if interface_name in interface_name_map:
            interface_cls_name = interface_name_map[interface_name]
            interface = socket.get_interface_object(interface_cls_name)

        elif allow_missing:
            interface = None

        else:
            errStr = ("Interface {} not found for socket " "class {}").format(
                interface_name, socket_cls_name
            )
            raise ValueError(errStr)

        return interface

    def _find_providing_interface(
        self, core, socket_cls_name, allow_missing=False, allow_multiple=False
    ):
        interfaces = self._find_providing_interfaces(core, socket_cls_name)
        interface = self._select_interface(
            interfaces, socket_cls_name, allow_missing, allow_multiple
        )

        return interface

    def _find_receiving_interface(
        self, core, socket_cls_name, allow_missing=False, allow_multiple=False
    ):
        interfaces = self._find_receiving_interfaces(core, socket_cls_name)
        interface = self._select_interface(
            interfaces, socket_cls_name, allow_missing, allow_multiple
        )

        return interface

    def _find_providing_interfaces(self, core: Core, socket_cls_name):
        socket = core.socket_map[socket_cls_name]
        providers = socket.get_providing_interfaces(self._id)

        interfaces = self._get_socket_interfaces(socket, providers)

        return interfaces

    def _find_receiving_interfaces(self, core: Core, socket_cls_name):
        socket = core.socket_map[socket_cls_name]
        receivers = socket.get_receiving_interfaces(self._id)
        interfaces = self._get_socket_interfaces(socket, receivers)

        return interfaces

    def _get_providers(
        self,
        core,
        interface_name,
        auto_interface_name=None,
        dict_value_attr=None,
    ):
        providers = self._find_providing_interfaces(core, interface_name)

        if auto_interface_name is not None:
            auto_providers = self._find_providing_interfaces(
                core, auto_interface_name
            )
            providers.extend(auto_providers)

        provider_names = [x.get_name() for x in providers]

        if dict_value_attr is None:
            return provider_names

        provider_values = []

        for provider in providers:
            provider_attr = getattr(provider, dict_value_attr)
            provider_values.append(provider_attr())

        result = {
            name: value for name, value in zip(provider_names, provider_values)
        }

        return result

    def _get_receivers(
        self,
        core,
        project,
        interface_name,
        auto_interface_name=None,
        dict_value_attr=None,
    ):
        receivers = self._find_receiving_interfaces(core, interface_name)
        active_receivers = []

        for interface in receivers:
            if core.can_load_interface(project, interface, self._id):
                active_receivers.append(interface)

        if auto_interface_name is not None:
            receivers = self._find_receiving_interfaces(
                core, auto_interface_name
            )

            for interface in receivers:
                if core.can_load_interface(project, interface, self._id):
                    active_receivers.append(interface)

        active_receiver_names = [x.get_name() for x in active_receivers]

        if dict_value_attr is None:
            result = active_receiver_names
            return result

        receiver_values = []

        for receiver in active_receivers:
            receiver_attr = getattr(receiver, dict_value_attr)
            receiver_values.append(receiver_attr())

        result = {
            name: value
            for name, value in zip(active_receiver_names, receiver_values)
        }

        return result

    def _get_providing_interface(
        self,
        core,
        super_interface_name,
        auto_interface_name=None,
        interface_name=None,
    ):
        interface = None

        if interface_name is not None:
            if interface_name in self._get_providers(
                core, super_interface_name
            ):
                interface = self._get_named_interface(
                    core, super_interface_name, interface_name
                )

            elif (
                auto_interface_name is not None
                and interface_name
                in self._get_providers(core, auto_interface_name)
            ):
                interface = self._get_named_interface(
                    core, auto_interface_name, interface_name
                )

            else:
                errStr = (
                    "Interface {} is not available for " "variable {}."
                ).format(interface_name, self._id)
                raise ValueError(errStr)

        if interface is None and auto_interface_name is not None:
            interface = self._find_providing_interface(
                core, auto_interface_name, allow_missing=True
            )

        if interface is None:
            interface = self._find_providing_interface(
                core,
                super_interface_name,
                allow_missing=True,
                allow_multiple=True,
            )

        return interface

    def _get_receiving_interface(
        self,
        core,
        project,
        super_interface_name,
        auto_interface_name=None,
        interface_name=None,
    ):
        interface = None

        if interface_name is not None:
            if interface_name in self._get_receivers(
                core, project, super_interface_name
            ):
                interface = self._get_named_interface(
                    core, super_interface_name, interface_name
                )

            elif (
                auto_interface_name is not None
                and interface_name
                in self._get_receivers(core, project, auto_interface_name)
            ):
                interface = self._get_named_interface(
                    core, auto_interface_name, interface_name
                )

            else:
                errStr = (
                    "Interface {} is not available for " "variable {}."
                ).format(interface_name, self._id)
                raise ValueError(errStr)

        if interface is None and auto_interface_name is not None:
            interface = self._find_receiving_interface(
                core, auto_interface_name, allow_missing=True
            )

        if interface is None:
            interface = self._find_receiving_interface(
                core,
                super_interface_name,
                allow_missing=True,
                allow_multiple=True,
            )

        return interface

    def _select_interface(
        self,
        interfaces,
        socket_cls_name,
        allow_missing=False,
        allow_multiple=False,
    ):
        if len(interfaces) == 0 and allow_missing:
            return None

        elif len(interfaces) == 0:
            errStr = (
                "No interfaces found of type {} for variable " "{}."
            ).format(socket_cls_name, self._id)
            raise IOError(errStr)

        elif not allow_multiple and len(interfaces) > 1:
            interface_strs = [x.get_name() for x in interfaces]
            infoStr = ", ".join(interface_strs)
            errStr = (
                "Multiple interfaces of type {} found for variable {}: " "{}."
            ).format(socket_cls_name, self._id, infoStr)
            raise ValueError(errStr)

        interface = interfaces[0]

        return interface

    def _write_interface(self, core, project, interface):
        if not core.can_load_interface(project, interface):
            errStr = (
                "The inputs to interface {} are not " "satisfied."
            ).format(interface.get_name())
            raise ValueError(errStr)

        interface = core.load_interface(project, interface)
        core.connect_interface(project, interface)

    @classmethod
    def _get_socket_interfaces(cls, socket, interface_cls_names):
        interfaces = []

        for interface_cls_name in interface_cls_names:
            interface = socket.get_interface_object(interface_cls_name)
            interfaces.append(interface)

        return interfaces


class InputVariable(Variable):
    """Class to indentify a particular input varible in the pipeline and set
    interfaces for it.
    """

    def get_interface(self):
        return self._interface

    def get_file_input_interfaces(self, core, include_auto=False):
        kargs = {
            "core": core,
            "interface_name": "FileInputInterface",
            "dict_value_attr": "get_valid_extensions",
        }
        if include_auto:
            kargs["auto_interface_name"] = "AutoFileInput"

        return self._get_providers(**kargs)

    def get_raw_interfaces(self, core, include_auto=False):
        args = [core, "RawInterface"]
        if include_auto:
            args.append("AutoRaw")

        return self._get_providers(*args)

    def get_query_interfaces(self, core, include_auto=False):
        args = [core, "QueryInterface"]
        if include_auto:
            args.append("AutoQuery")

        return self._get_providers(*args)

    def set_raw_interface(self, core, value, interface_name=None):
        interface = self._get_providing_interface(
            core, "RawInterface", "AutoRaw", interface_name
        )

        if interface is None:
            errStr = ("No raw input interface found for " "variable ").format()
            raise RuntimeError(errStr)

        interface.set_variables({self._id: value})
        self._interface = interface

    def set_file_interface(
        self,
        core,
        file_path: StrOrPath,
        interface_name=None,
    ):
        interface = self._get_providing_interface(
            core, "FileInputInterface", "AutoFileInput", interface_name
        )

        if interface is None:
            errStr = ("No file input interface found for " "variable ").format()
            raise RuntimeError(errStr)

        assert isinstance(interface, FileInterface)
        interface.set_file_path(file_path)
        self._interface = interface

    def set_query_interface(self, core, project, interface_name=None):
        # Check the status of the database.
        if project.get_database_credentials() is None:
            errStr = "No database credentials found"
            raise RuntimeError(errStr)

        interface = self._get_query_interface(core, project, interface_name)

        if interface is None:
            errStr = (
                "No database access interface found for " "variable "
            ).format()
            raise RuntimeError(errStr)

        self._interface = interface

    def read(
        self,
        core,
        project,
        overwrite=True,
        set_auto=False,
        log_exceptions=False,
    ):
        """Read data from the variable and inject the data into the active data
        state"""

        _read_variables(
            core, project, [self], overwrite, set_auto, log_exceptions
        )

    def read_auto(self, core, project, overwrite=True, log_exceptions=True):
        """Calls read with set_auto set to True."""

        self.read(core, project, overwrite, True, log_exceptions)

    def read_file(
        self,
        core,
        project,
        file_path: StrOrPath,
        file_interface=None,
        overwrite=True,
    ):
        self.set_file_interface(core, file_path, file_interface)
        self.read(core, project, overwrite)

    def _get_query_interface(self, core, project, interface_name=None):
        interface = self._get_providing_interface(
            core, "QueryInterface", "AutoQuery", interface_name
        )

        if interface is not None:
            if core.can_load_interface(project, interface):
                interface = core.load_interface(project, interface)
            else:
                interface = None

        if interface is None:
            return None

        metadata = self.get_metadata(core)

        # Allow the auto_only meta data variable to restrict execution of the
        # auto recovery.
        if (
            not metadata.auto_only
            or metadata.auto_only
            and core.has_data(project, metadata.auto_only)
        ):
            result = interface

        else:
            result = None

        return result


class OutputVariable(Variable):
    """Class to to indentify for a particular output varible in the pipeline"""


### MODULE FUNCTIONS


def set_output_scope(
    core,
    project,
    scope=None,
    sim_index=None,
    sim_title=None,
    default_scope="global",
):
    simulation = project.get_simulation(sim_index, sim_title)

    if scope is None:
        scope = simulation.output_scope

    if scope is None:
        scope = default_scope

    # Mask either global or local theme results
    if scope == "local":
        unmask_marker = core._markers["local"]
    elif scope == "global":
        unmask_marker = core._markers["global"]
    else:
        errStr = "Theme output scope '{}' is not valid".format(scope)
        raise ValueError(errStr)

    # Record the scope
    simulation.output_scope = scope

    # Get the level root from the inspection level
    interface_name = simulation.level_map[simulation.get_inspection_level()]

    if interface_name is None:
        return

    level_root = simulation.level_map[simulation.get_inspection_level()].lower()

    unmask_level = "{} {}".format(level_root, unmask_marker)

    # Mask all the local and global output levels
    core.mask_states(
        project,
        simulation,
        core._markers["local"],
        no_merge=True,
        update_status=False,
    )
    core.mask_states(
        project,
        simulation,
        core._markers["global"],
        no_merge=True,
        update_status=False,
    )

    # Unmask the levels relevant to the branch and scope.
    core.unmask_states(project, simulation, unmask_level, update_status=False)

    core.set_interface_status(project, simulation)


def _get_connector(project, hub_name):
    simulation = project.get_simulation()
    hub_names = simulation.get_hub_ids()

    if hub_name not in hub_names:
        errStr = ('Hub name "{}" is not contained in the ' "project.").format(
            hub_name
        )
        raise ValueError(errStr)

    result = Connector(hub_name)

    return result


def _can_read_value(core, project, variable_id, overwrite=True):
    result = True

    if not overwrite and core.has_data(project, variable_id):
        result = False

    return result


def _get_read_values(core, project, interface, variable_ids):
    if interface is None:
        errStr = "No interface set. Aborting."
        raise ValueError(errStr)

    variable_values = []

    # Test that the variables are in the data catalog
    for variable_id in variable_ids:
        core.check_valid_variable(variable_id)

    # Get the value for the variables from the interface
    interface = core.connect_interface(project, interface)

    for variable_id in variable_ids:
        value = interface.get_data(variable_id)
        variable_values.append(value)

    return variable_values


def _read_variables(
    core,
    project,
    variables: Sequence[InputVariable],
    overwrite=True,
    set_auto=False,
    log_exceptions=False,
):
    """Read data from the given list of variables and inject the data
    into the active data state"""

    results = {}
    variables = list(variables)

    while variables:
        variable = variables.pop()

        if not _can_read_value(core, project, variable._id, overwrite):
            continue

        if set_auto:
            if project.get_database_credentials() is None:
                continue

            interface = variable._get_query_interface(core, project)

            if interface is None:
                continue

        else:
            interface = variable._interface

        fetch_var_ids = [variable._id]
        all_var_ids = [x._id for x in variables]

        # Check if additional variables are provided by the same interface
        needed = set(all_var_ids)
        outputs = interface.declare_outputs()

        if outputs is None:
            provided = set()
        else:
            provided = set(outputs)

        add_var_ids = needed & provided

        for add_var_id in add_var_ids:
            # Remove the variable from the variables list
            var_dex = all_var_ids.index(add_var_id)
            add_var = variables.pop(var_dex)

            if add_var._interface is None:
                add_var_interface_name = None
            else:
                add_var_interface_name = add_var._interface.get_name()

            variable_interface_name = interface.get_name()

            # If this is happening without auto_queries and different
            # interfaces were set issue a warning
            if (
                not set_auto
                and add_var_interface_name is not None
                and variable_interface_name != add_var_interface_name
            ):
                msg = (
                    "Variable {} is being retrieved from interface {} "
                    "rather than {}"
                ).format(
                    add_var_id, add_var_interface_name, variable_interface_name
                )
                module_logger.warning(msg)

            if not _can_read_value(core, project, add_var_id, overwrite):
                continue

            fetch_var_ids.append(add_var_id)

        varsStr = ", ".join(fetch_var_ids)
        msg = "Reading variables {}".format(varsStr)
        module_logger.debug(msg)

        if log_exceptions:
            try:
                values = _get_read_values(
                    core, project, interface, fetch_var_ids
                )

            except:  # noqa: E722
                e = sys.exc_info()[0]
                msg = ("Reading variables {} generated error: " "{}").format(
                    varsStr, e
                )
                logging.exception(msg)

                continue

        else:
            values = _get_read_values(core, project, interface, fetch_var_ids)

        for var_id, value in zip(fetch_var_ids, values):
            results[var_id] = value

    if not results:
        return

    # Add data to the cores active data state
    core.add_datastate(
        project,
        identifiers=results.keys(),
        values=results.values(),
        log_exceptions=log_exceptions,
    )



#    Copyright (C) 2016-2024 Mathew Topper
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
from collections import OrderedDict
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import colormaps
from mdo_engine.utilities.plugins import Plugin

import dtocean_plugins.strategies as strategies
import dtocean_plugins.tools as tools
from dtocean_core.core import Core, Project
from dtocean_plugins.strategies.base import Strategy
from dtocean_plugins.tools.base import Tool

from .menu import ModuleMenu

# Set up logging
module_logger = logging.getLogger(__name__)


class ExtensionManager(Plugin):
    """Extension framework discovery"""

    def __init__(self, module, cls_name):
        super(ExtensionManager, self).__init__()
        self._plugin_classes = self._discover_classes(module, cls_name)
        self._plugin_names = self._discover_names()

    def get_available(self) -> list[str]:
        plugin_names = self._plugin_names.keys()
        return list(plugin_names)

    def _get_plugin(self, plugin_name):
        if plugin_name not in self.get_available():
            errStr = ("Name {} is not a recognised " "plugin").format(
                plugin_name
            )
            raise KeyError(errStr)

        cls_name = self._plugin_names[plugin_name]
        ExtensionCls = self._plugin_classes[cls_name]
        ext_obj = ExtensionCls()

        return ext_obj

    def _discover_classes(self, module, cls_name):
        """Retrieve all of the available plugin classes"""

        log_msg = "Searching for {} classes".format(cls_name)
        module_logger.debug(log_msg)

        cls_map = self._discover_plugins(module, cls_name, warn_import=True)

        return cls_map

    def _discover_names(self):
        plugin_names = {}

        # Work through the interfaces
        for cls_name, cls_attr in self._plugin_classes.items():
            name = cls_attr.get_name()
            plugin_names[name] = cls_name

        return plugin_names


class StrategyManager(ExtensionManager):
    """Strategy discovery"""

    def __init__(self, module=strategies, cls_name="Strategy"):
        super(StrategyManager, self).__init__(module, cls_name)

        self._module_menu = ModuleMenu()

    def get_strategy(self, strategy_name) -> Strategy:
        if strategy_name not in self.get_available():
            errStr = ("Name {} is not a recognised " "strategy").format(
                strategy_name
            )
            raise KeyError(errStr)

        strategy_obj = self._get_plugin(strategy_name)

        return strategy_obj

    def get_level_values(
        self,
        core,
        project,
        var_id,
        strategy=None,
        sim_titles=None,
        scope="global",
    ):
        if scope not in ["global", "local"]:
            errStr = (
                "Argument 'scope' must have value 'global' or 'local', "
                "not {}"
            ).format(scope)
            raise ValueError(errStr)

        chosen_modules = self._module_menu.get_active(core, project)
        output_levels = [
            "{} {} output".format(x, scope).lower() for x in chosen_modules
        ]

        # Mask all the global and local output levels before getting the
        # required output level
        force_masks = ["local", "global"]

        if strategy is not None:
            sim_titles = strategy.get_simulation_record()

        if sim_titles is None:
            sim_indexes = range(len(project))
            sim_titles = [
                project.get_simulation_title(index=x) for x in sim_indexes
            ]

        else:
            sim_indexes = project.get_simulation_indexes(
                sim_titles, raise_if_missing=False
            )

        sim_levels = OrderedDict()

        for sim_title, sim_index in zip(sim_titles, sim_indexes):
            if sim_index is None:
                continue
            if sim_title is None:
                sim_title = sim_index

            level_values = core.get_level_values(
                project, var_id, output_levels, force_masks, sim_index=sim_index
            )

            sim_levels[sim_title] = level_values

        return sim_levels

    def get_level_values_df(
        self,
        core,
        project,
        var_id,
        strategy=None,
        sim_titles=None,
        scope="global",
    ):
        sim_levels = self.get_level_values(
            core, project, var_id, strategy, sim_titles, scope
        )
        done_levels = self._module_menu.get_completed(core, project)

        # Check the number of levels in each simulation
        sim_lengths = []

        for level_values in sim_levels.values():
            sim_lengths.append(len(level_values.values()))

        level_set = set(sim_lengths)

        if len(level_set) != 1:
            errStr = "The number of levels in each simulation is not equal"
            raise ValueError(errStr)

        sim_names = []
        level_lists = {k: [] for k in done_levels}

        for sim_key, level_values in sim_levels.items():
            sim_names.append(sim_key)

            for name in done_levels:
                find_levels = [
                    level for level in level_values if name.lower() in level
                ]

                if len(find_levels) > 1:
                    errStr = (
                        "More than one level matches module name " "{}"
                    ).format(name)
                    raise RuntimeError(errStr)

                if len(find_levels) == 0:
                    value = np.nan
                else:
                    found_level = find_levels[0]
                    value = level_values[found_level]

                level_lists[name].append(value)

        raw_dict = {"Simulation Name": sim_names}
        raw_dict.update(level_lists)

        raw_cols = ["Simulation Name"]
        raw_cols.extend(done_levels)

        df = pd.DataFrame(raw_dict)
        df = df[raw_cols]

        return df

    def get_level_values_plot(
        self,
        core,
        project,
        var_id,
        strategy=None,
        sim_titles=None,
        scope="global",
        legend_loc="upper left",
        max_lines=10,
    ):
        sim_levels = self.get_level_values(
            core, project, var_id, strategy, sim_titles, scope
        )
        done_levels = self._module_menu.get_completed(core, project)

        # Check the number of levels in each simulation
        sim_lengths = []

        for level_values in sim_levels.values():
            sim_lengths.append(len(level_values.values()))

        level_set = set(sim_lengths)

        if len(level_set) != 1:
            errStr = "The number of levels in each simulation is not equal"
            raise ValueError(errStr)

        fig = plt.figure()
        ax = fig.gca()

        if len(sim_levels) < 10:
            num_plots = len(sim_levels)
        else:
            num_plots = 10

        colormap = colormaps["Set1"]
        colors = [colormap(i) for i in np.linspace(0, 1, num_plots)]
        ax.set_prop_cycle(color=colors)

        x = range(len(done_levels))
        metadata = core.get_metadata(var_id)

        for i, (sim_key, level_values) in enumerate(sim_levels.items()):
            # Do not exceed the maximum number of lines
            if i == max_lines:
                break

            sane_values = []

            for name in done_levels:
                find_levels = [
                    level for level in level_values if name.lower() in level
                ]

                if len(find_levels) > 1:
                    errStr = (
                        "More than one level matches module name " "{}"
                    ).format(name)
                    raise RuntimeError(errStr)

                if len(find_levels) == 0:
                    value = np.nan
                else:
                    found_level = find_levels[0]
                    value = level_values[found_level]

                sane_values.append(value)

            plt.plot(x, sane_values, "-o", label=sim_key)
            plt.ylabel(metadata.title)

        plt.xticks(x, done_levels, rotation="vertical")

        y_label = metadata.title

        if metadata.units is not None:
            y_label = "{} ({})".format(y_label, metadata.units[0])

        plt.ylabel(y_label)
        plt.legend(loc=legend_loc)
        plt.title("Module Comparison ({} Scope)".format(scope.capitalize()))

        plt.tight_layout()

        return plt.gcf()

    def get_comparison_values(
        self,
        core,
        project,
        var_one_id,
        var_two_id,
        module=None,
        strategy=None,
        scope="global",
        sort=True,
    ):
        if scope not in ["global", "local"]:
            errStr = (
                "Argument 'scope' must have value 'global' or 'local', "
                "not {}"
            ).format(scope)
            raise ValueError(errStr)

        all_modules = self._module_menu.get_active(core, project)

        # Determine at which module to carry out the comparison
        if module is None:
            module = all_modules[-1]

        elif module not in all_modules:
            errStr = (
                "Module '{}' is not in the list of active " "modules"
            ).format(module)
            raise ValueError(errStr)

        output_level = "{} {} output".format(module, scope).lower()

        # If a strategy is given then just use its simulation indexes
        if strategy is None:
            sim_indexes = None
        else:
            sim_titles = strategy.get_simulation_record()
            sim_indexes = project.get_simulation_indexes(sim_titles)

        var_one_values = core.get_project_values(
            project,
            var_one_id,
            output_level,
            force_indexes=sim_indexes,
            allow_none=True,
        )

        var_two_values = core.get_project_values(
            project,
            var_two_id,
            output_level,
            force_indexes=sim_indexes,
            allow_none=True,
        )

        if var_one_values is None or var_two_values is None:
            x = []
            y = []
        else:
            x = [v for (n, v) in var_one_values]
            y = [v for (n, v) in var_two_values]

        if not sort:
            return x, y

        # Sort by the x value
        points = zip(x, y)
        sorted_points = sorted(points)
        new_x = [point[0] for point in sorted_points]
        new_y = [point[1] for point in sorted_points]

        return new_x, new_y

    def get_comparison_values_df(
        self,
        core,
        project,
        var_one_id,
        var_two_id,
        module=None,
        strategy=None,
        scope="global",
    ):
        # Get the comparison values
        x, y = self.get_comparison_values(
            core,
            project,
            var_one_id,
            var_two_id,
            module=module,
            strategy=strategy,
            scope=scope,
        )

        # Redetermine the module used
        if module is None:
            all_modules = self._module_menu.get_active(core, project)
            module = all_modules[-1]

        var_one_meta = core.get_metadata(var_one_id)
        var_two_meta = core.get_metadata(var_two_id)

        var_one_str = var_one_meta.title

        if var_one_meta.units is not None:
            var_one_str = "{} ({})".format(var_one_str, var_one_meta.units[0])

        var_two_str = var_two_meta.title

        if var_two_meta.units is not None:
            var_two_str = "{} ({})".format(var_two_str, var_two_meta.units[0])

        raw_dict = {var_one_str: x, var_two_str: y}
        raw_cols = [var_one_str, var_two_str]

        df = pd.DataFrame(raw_dict)
        df = df[raw_cols]

        return df

    def get_comparison_values_plot(
        self,
        core,
        project,
        var_one_id,
        var_two_id,
        module=None,
        strategy=None,
        scope="global",
    ):
        # Get the comparison values
        x, y = self.get_comparison_values(
            core,
            project,
            var_one_id,
            var_two_id,
            module=module,
            strategy=strategy,
            scope=scope,
        )

        # Convert any string x-values to a numerical range and add ticks later
        x_ticks = None

        if any(isinstance(i, str) for i in x):
            x_ticks = x
            x = range(len(x))

        plt.figure()

        # Redetermine the module used
        if module is None:
            all_modules = self._module_menu.get_active(core, project)
            module = all_modules[-1]

        var_one_meta = core.get_metadata(var_one_id)
        var_two_meta = core.get_metadata(var_two_id)

        var_one_str = var_one_meta.title

        if var_one_meta.units is not None:
            var_one_str = "{} ({})".format(var_one_str, var_one_meta.units[0])

        var_two_str = var_two_meta.title

        if var_two_meta.units is not None:
            var_two_str = "{} ({})".format(var_two_str, var_two_meta.units[0])

        if x_ticks is None:
            if len(x) == 1:
                plt.plot(x, y, marker="o", ls="")
            else:
                plt.plot(x, y)

        else:
            plt.bar(x, y, align="center")
            plt.xticks(x, x_ticks, rotation="vertical")

        plt.xlabel(var_one_str)
        plt.ylabel(var_two_str)
        plt.title(
            "Simulation Comparison at Module: {} " "({} Scope)".format(
                module, scope.capitalize()
            ),
            y=1.08,
        )

        plt.tight_layout()

        return plt.gcf()

    def dump_strategy(self, strategy: Strategy, dump_path):
        if os.path.splitext(dump_path)[1] != ".json":
            errStr = "Argument dump_path must be a file with .json extension"
            raise ValueError(errStr)

        stg_dict = self._get_dump_dict(strategy)

        with open(dump_path, "w") as fstream:
            json.dump(stg_dict, fstream)

    def load_strategy(self, load_path) -> Strategy:
        # OK need to consider if we have a json file
        if not os.path.isfile(load_path) and ".json" not in load_path:
            errStr = "Argument load_path must be a file with .json extension"
            raise ValueError(errStr)

        # Load the strategy file
        with open(load_path, "r") as fstream:
            stg_dict = json.load(fstream)

        # Now deserialise the data
        new_strategy = self._set_load_dict(stg_dict)

        return new_strategy

    def _get_dump_dict(self, strategy: Strategy):
        # Now store the strategy information
        stg_name_str = strategy.get_name()

        stg_dict = {
            "name": stg_name_str,
            "sim_record": strategy._sim_record,
            "config": {
                "version": strategy.version,
                "data": strategy.dump_config(strategy._config),
            },
            "sim_details": strategy.dump_sim_details(strategy.sim_details),
            "version": 1,
        }

        return stg_dict

    def _set_load_dict(self, stg_dict):
        if stg_dict["version"] != 1:
            raise RuntimeError("Data version not recognised")

        # Now build the strategy
        new_strategy = self.get_strategy(stg_dict["name"])

        # Now deserialise the data
        sim_titles = stg_dict["sim_record"]
        new_strategy._sim_record = sim_titles
        new_strategy._config = new_strategy.load_config(
            stg_dict["config"]["data"],
            stg_dict["config"]["version"],
        )
        new_strategy.sim_details = new_strategy.load_sim_details(
            stg_dict["sim_details"]
        )

        return new_strategy


class ToolManager(ExtensionManager):
    """Tool discovery and execution"""

    def __init__(self, module=tools, cls_name="Tool"):
        super(ToolManager, self).__init__(module, cls_name)

    def get_tool(self, tool_name) -> Tool:
        if tool_name not in self.get_available():
            errStr = ("Name {} is not a recognised " "tool").format(tool_name)
            raise KeyError(errStr)

        tool_obj = self._get_plugin(tool_name)

        return tool_obj

    def can_execute_tool(
        self,
        core: Core,
        project: Optional[Project],
        tool: Tool,
    ) -> bool:
        if project is None:
            return False

        result = False

        if core.can_load_interface(project, tool):
            result = True

        return result

    def execute_tool(self, core: Core, project: Project, tool: Tool):
        if not core.can_load_interface(project, tool):
            errStr = ("The inputs to tool {} are not " "satisfied.").format(
                tool.get_name()
            )
            raise ValueError(errStr)

        interface = core.load_interface(project, tool)
        core.connect_interface(project, interface)
